import { createStore as createStore$1 } from 'redux';
import produce from 'immer';

function isPlainObject(obj) {
    if (typeof obj !== 'object' || obj === null)
        return false;
    var proto = obj;
    while (Object.getPrototypeOf(proto) !== null) {
        proto = Object.getPrototypeOf(proto);
    }
    return Object.getPrototypeOf(obj) === proto;
}
const isPrimitive = (v) => typeof v === 'boolean' || typeof v === 'number' || typeof v === 'string' || v === null;
const isValidPayload = (v) => Array.isArray(v) || isPlainObject(v) || isPrimitive(v);

const wrongModify = '[Imdux] Do not modify value of store directly, use dispatch instead.';
const payloadNotValid = '[Imdux] Payload should always be a plain object, other value will be ignored.';

class Store {
    constructor(states, refs, options) {
        this.Dispatch = {};
        this.State = {};
        const drafts = {};
        const initialStates = {};
        const root = {};
        Object.keys(states).forEach(key => {
            root[key] = states[key].reducers;
            initialStates[key] = states[key].initialState;
            const getState = (key) => this.redux.getState()[key];
            Object.defineProperty(this.State, key, {
                get() {
                    return getState(key);
                },
                set() {
                    throw new Error(wrongModify);
                },
            });
        });
        const initReducers = (reducers, namespace, dispatch) => {
            Object.keys(reducers).forEach(key => {
                if (key.toLowerCase() === 'refs') {
                    throw new Error(`Cannot use reducer name 'refs'`);
                }
                const reducer = reducers[key];
                const path = namespace === '' ? key : namespace + '/' + key;
                if (typeof reducer === 'function') {
                    drafts[path] = (current, action) => produce(current, (draft) => reducer(action.payload, { draft }));
                    dispatch[key] = (payload) => {
                        if (!isValidPayload(payload)) {
                            console.warn(payloadNotValid);
                            return;
                        }
                        this.redux.dispatch({ type: path, payload });
                    };
                }
                else {
                    dispatch[key] = {};
                    initReducers(reducer, path, dispatch[key]);
                }
            });
        };
        initReducers(root, '', this.Dispatch);
        Object.keys(states).forEach(ns => {
        });
        const refsMap = {};
        const initRefs = (refs, prefix) => {
            Object.keys(refs).forEach(key => {
                const oldValue = refs[key];
                const path = `${prefix}/${key}`;
                if (oldValue === null) {
                    refsMap[path] = oldValue;
                    const dispatch = (key, action) => this.redux.dispatch({ type: `${path}/${action}` });
                    Object.defineProperty(refs, key, {
                        get() {
                            return refsMap[path];
                        },
                        set(newValue) {
                            if (newValue !== refsMap[path]) {
                                if (newValue !== null && refsMap[path] === null) {
                                    dispatch(key, 'mount');
                                }
                                else if (newValue === null && refsMap[path] !== null) {
                                    dispatch(key, 'unmount');
                                }
                                else if (newValue !== null && refsMap[path] !== null) {
                                    dispatch(key, 'update');
                                }
                                else {
                                    dispatch(key, 'unknown');
                                }
                                refsMap[path] = newValue;
                            }
                        },
                    });
                }
                else if (isPlainObject(oldValue)) {
                    initRefs(oldValue, path);
                }
                else {
                    throw new Error(`Not support ${typeof oldValue} type in refs.`);
                }
            });
        };
        initRefs(refs, 'refs');
        const rootReducer = (current, action) => {
            if (current === undefined) {
                return initialStates;
            }
            else if (drafts[action.type] !== undefined) {
                return drafts[action.type](current, action);
            }
            else {
                return current;
            }
        };
        this.redux = createStore$1(rootReducer, 
        // https://medium.com/@e_himmelfarb/implement-redux-devtools-extension-with-thunk-and-other-async-middleware-20e97100b2b0
        options?.devtool ? window.__REDUX_DEVTOOLS_EXTENSION__?.({ trace: true, name: options.name }) : (x) => x);
    }
}
function loop(services, Service, Dispatch, State) {
    Object.keys(services).forEach(key => {
        const service = services[key];
        if (typeof service === 'function') {
            services[key] = (...args) => service(...args, { Dispatch, Service, State });
        }
        else if (typeof service === 'object') {
            loop(service, Service, Dispatch, State);
        }
        else {
            throw new Error('Invalid service type.');
        }
    });
    return Service;
}
function createStore(refs, states, services, options) {
    const store = new Store(states, refs, options);
    const Dispatch = store.Dispatch;
    const State = store.State;
    const Service = loop(services, services, Dispatch, State);
    const Refs = refs;
    const getState = store.redux.getState;
    const subscribe = store.redux.subscribe;
    const redux = store.redux;
    return { Dispatch, Service, State, Refs, redux, getState, subscribe };
}

export { createStore };
