import * as Redux from 'redux';
import produce from 'immer';

import { isPlainObject, isValidPayload } from './utils';
import { payloadNotValid, wrongModify } from './message';
import { Imdux } from './type';

class Store {
    Dispatch: any;
    State: any;
    redux: Redux.Store<any, Redux.AnyAction>;

    constructor(states: Imdux.States, refs: any, options?: Imdux.Options) {
        this.Dispatch = {};
        this.State = {};

        const drafts: any = {};
        const reducers: any = {};
        const initialStates: any = {};
        const root: any = {};

        Object.keys(states).forEach(key => {
            root[key] = states[key].reducers;
            initialStates[key] = states[key].initialState;
            const getState = (key: string) => this.redux.getState()[key];
            Object.defineProperty(this.State, key, {
                get() {
                    return getState(key);
                },
                set() {
                    throw new Error(wrongModify);
                },
            });
        });

        const initReducers = (reducers: any, namespace: string, dispatch: any) => {
            Object.keys(reducers).forEach(key => {
                if (key.toLowerCase() === 'refs') {
                    throw new Error(`Cannot use reducer name 'refs'`);
                }
                const reducer = reducers[key];
                const path = namespace === '' ? key : namespace + '/' + key;
                if (typeof reducer === 'function') {
                    drafts[path] = (current: any, action: any) => produce(current, (draft: any) => reducer(action.payload, { draft }));
                    dispatch[key] = (payload: any) => {
                        if (!isValidPayload(payload)) {
                            console.warn(payloadNotValid);
                            return;
                        }
                        this.redux.dispatch({ type: path, payload });
                    };
                } else {
                    dispatch[key] = {};
                    initReducers(reducer, path, dispatch[key]);
                }
            });
        };

        initReducers(root, '', this.Dispatch);

        Object.keys(states).forEach(ns => {
            reducers[ns] = (current: any, action: any) => {
                const split: Array<string> = action.type.split('/');
                const namespace = split[0];
                if (namespace === ns && drafts[action.type]) {
                    return drafts[action.type](current, action);
                } else {
                    return current;
                }
            };
        });

        const refsMap: any = {};
        const initRefs = (refs: any, prefix: string) => {
            Object.keys(refs).forEach(key => {
                const oldValue = refs[key];
                const path = `${prefix}/${key}`;
                if (oldValue === null) {
                    refsMap[path] = oldValue;
                    const dispatch = (key: string, action: string) => this.redux.dispatch({ type: `${path}/${action}` });
                    Object.defineProperty(refs, key, {
                        get() {
                            return refsMap[path];
                        },
                        set(newValue: any) {
                            if (newValue !== refsMap[path]) {
                                if (newValue !== null && refsMap[path] === null) {
                                    dispatch(key, 'mount');
                                } else if (newValue === null && refsMap[path] !== null) {
                                    dispatch(key, 'unmount');
                                } else if (newValue !== null && refsMap[path] !== null) {
                                    dispatch(key, 'update');
                                } else {
                                    dispatch(key, 'unknown');
                                }
                                refsMap[path] = newValue;
                            }
                        },
                    });
                } else if (isPlainObject(oldValue)) {
                    initRefs(oldValue, path);
                } else {
                    throw new Error(`Not support ${typeof oldValue} type in refs.`);
                }
            });
        };

        initRefs(refs, 'refs');

        const rootReducer = (current: any, action: any) => {
            if (current === undefined) {
                return initialStates;
            } else if (drafts[action.type] !== undefined) {
                return drafts[action.type](current, action);
            } else {
                return current;
            }
        };

        this.redux = Redux.createStore(
            rootReducer,
            // https://medium.com/@e_himmelfarb/implement-redux-devtools-extension-with-thunk-and-other-async-middleware-20e97100b2b0
            options?.devtool ? (window as any).__REDUX_DEVTOOLS_EXTENSION__?.({ trace: true, name: options.name }) : (x: any) => x,
        );
    }
}

function loop(services: any, Service: any, Dispatch: any, State: any) {
    Object.keys(services).forEach(key => {
        const service = services[key];
        if (typeof service === 'function') {
            services[key] = (...args: any[]) => service(...args, { Dispatch, Service, State });
        } else if (typeof service === 'object') {
            loop(service, Service, Dispatch, State);
        } else {
            throw new Error('Invalid service type.');
        }
    });
    return Service;
}

export function createStore<R, T extends Imdux.States, S extends Imdux.Services>(refs: R, states: T, services: S, options?: Imdux.Options) {
    const store: Imdux.Store<T> = new Store(states, refs, options);
    const Dispatch = store.Dispatch;
    const State = store.State;
    const Service: S = loop(services, services, Dispatch, State);
    const Refs = refs;
    const getState = store.redux.getState;
    const subscribe = store.redux.subscribe;
    const redux = store.redux;
    return { Dispatch, Service, State, Refs, redux, getState, subscribe };
}
